package de.unikoblenz.fgbks.base.builder;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;

/** @param <E> The type of the object being constructed. */
public class DefaultBuilder<E> extends AbstractBuilder<E, DefaultBuilder<E>> {

  protected DefaultBuilder() {
    super();
  }

  protected DefaultBuilder(E value) {
    super(value);
  }

  /**
   * Liefert E so zurück
   *
   * @return DefaultBuilder
   */
  @Override
  protected DefaultBuilder<E> thisBuilder() {
    return this;
  }

  @Override
  protected void validate() {
    // nothing to validate
  }

  /**
   * @return <tt>new E()</tt>, where E is the type of the object being constructed which is the type
   *     parameter of this class.
   * @throws IllegalStateException
   *     <ul>
   *       <li>when the type parameter cannot be extracted or is no {@link Class}, but a generic
   *           type
   *       <li>or when no default-constructor can be found
   *       <li>or when the constructor throws an exception other than {@link RuntimeException}
   *     </ul>
   */
  @Override
  protected E newInstance() {
    Class<?> directSubclass = getClass();
    while (directSubclass.getSuperclass() != DefaultBuilder.class) {
      directSubclass = directSubclass.getSuperclass();
    }
    Type genericSuperclass = directSubclass.getGenericSuperclass();
    if (!(genericSuperclass instanceof ParameterizedType)) {
      throw new IllegalStateException(
          "Generic type argument E missing for superclass " + DefaultBuilder.class.getSimpleName());
    }
    ParameterizedType parameterizedSuperclass = (ParameterizedType) genericSuperclass;
    Type valueType = parameterizedSuperclass.getActualTypeArguments()[0];
    if (!(valueType instanceof Class)) {
      throw new IllegalStateException(
          "No generics allowed for type argument, but found " + valueType);
    }
    Class<E> type = (Class<E>) valueType;
    try {
      Constructor<E> constructor = type.getDeclaredConstructor();
      constructor.setAccessible(true);
      return constructor.newInstance();
    } catch (NoSuchMethodException e) {
      throw new IllegalStateException("No default-constructor found for " + type.getName(), e);
    } catch (InvocationTargetException e) {
      if (e.getTargetException() instanceof RuntimeException) {
        throw (RuntimeException) e.getTargetException();
      } else {
        throw new IllegalStateException(e.getTargetException());
      }
    } catch (ReflectiveOperationException e) {
      throw new IllegalStateException(e);
    }
  }
}
