package de.unikoblenz.fgbks.dmn.core.verifier.helper;

import static de.unikoblenz.fgbks.dmn.core.verifier.helper.Boundary.BoundType.EXCLUSIVE;
import static de.unikoblenz.fgbks.dmn.core.verifier.helper.Boundary.BoundType.INCLUSIVE;

import de.unikoblenz.fgbks.base.builder.DefaultBuilder;
import java.util.Objects;
import java.util.Optional;

/** Class for manage boundaries and intervals. Can also handle String values. */
public class Boundary implements Comparable<Boundary> {

  public enum BoundType {
    INCLUSIVE,
    EXCLUSIVE;

    public BoundType getOp() {
      return this == INCLUSIVE ? EXCLUSIVE : INCLUSIVE;
    }

    public static BoundType getOp(BoundType boundType) {
      return boundType.getOp();
    }
  }

  private double lowerBound;
  private BoundType lowerBoundType;
  private double upperBound;
  private BoundType upperBoundType;
  private boolean integerValue;

  private Boundary() {
    super();
    this.integerValue = false;
  }

  public double getLowerBound() {
    return lowerBound;
  }

  public BoundType getLowerBoundType() {
    return lowerBoundType;
  }

  public double getUpperBound() {
    return upperBound;
  }

  public BoundType getUpperBoundType() {
    return upperBoundType;
  }

  private void setLowerBound(double lowerBound) {
    this.lowerBound = lowerBound;
  }

  private void setLowerBoundType(BoundType lowerBoundType) {
    this.lowerBoundType = lowerBoundType;
  }

  private void setUpperBound(double upperBound) {
    this.upperBound = upperBound;
  }

  private void setUpperBoundType(BoundType upperBoundType) {
    this.upperBoundType = upperBoundType;
  }

  public boolean isIntegerValue() {
    return integerValue;
  }

  private void setIntegerValue(boolean integerValue) {
    this.integerValue = integerValue;
  }

  /**
   * Check, if a number is in between the current boundary
   *
   * @return true or false
   */
  public boolean isNumberInRange(double number) {
    if (number < lowerBound && lowerBound != Double.MIN_VALUE
        || number == Double.MIN_VALUE && lowerBound != Double.MIN_VALUE) {
      return false;
    }
    if (number > upperBound) {
      return false;
    }
    if (number == lowerBound && lowerBoundType == EXCLUSIVE) {
      return false;
    }
    if (number == upperBound && upperBoundType == EXCLUSIVE) {
      return false;
    }
    return true;
  }

  /**
   * Check, if a number is in between the current boundary
   *
   * @return true or false
   */
  public boolean isNumberInRange(double number, BoundType boundType) {
    if ((number < lowerBound && lowerBound != Double.MIN_VALUE || number > upperBound)) {
      return false;
    } else if (boundType == INCLUSIVE
        || ((number > lowerBound || lowerBound == Double.MIN_VALUE) && number < upperBound)) {
      return isNumberInRange(number);
    } else {
      return false;
    }
  }

  /**
   * Check, if a other Boundary is overlapping (NOT equals and NO subsumption)
   *
   * <p>Example:
   *
   * <blockquote>
   *
   * <pre>
   *    true: (overlapping)
   *     o------o
   *        o-----o
   *    false: (not overlapping)
   *     o---o
   *          o--o
   *    false: (subsumption)
   *      o--o
   *       o-o
   *    false: (equals)
   *      o---o
   *      o---o
   *    false: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  public boolean isOverlapping(Boundary other) {
    return !isEqualsTo(other) && !isNotInContact(other) && !isSubsumption(other);
  }

  /**
   * Check, if a other Boundary is a subsumption (NOT equals and NO overlapping) True, if this
   * subsumes other or other subsumes this.
   *
   * <p>Example:
   *
   * <blockquote>
   *
   * <pre>
   *    false: (overlapping)
   *     o------o
   *        o-----o
   *    false: (not overlapping)
   *     o---o
   *          o--o
   *    true: (subsumption)
   *      o--o
   *       o-o
   *    false: (equals)
   *      o---o
   *      o---o
   *    true: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  public boolean isSubsumption(Boundary other) {
    if (equals(other)) {
      return false;
    }

    if (other.lowerBound == this.lowerBound
            && (other.lowerBoundType == EXCLUSIVE && this.lowerBoundType == EXCLUSIVE)
        || (this.upperBound == other.upperBound
            && (this.upperBoundType == EXCLUSIVE && other.upperBoundType == EXCLUSIVE))) {
      return true;
    }

    if (this.lowerBound == other.lowerBound) {
      if (this.lowerBoundType == EXCLUSIVE) {
        return other.isNumberInRange(upperBound);
      }
      if (other.lowerBoundType == EXCLUSIVE) {
        return isNumberInRange(other.upperBound);
      }
    }

    if (this.upperBound == other.upperBound) {
      if (this.upperBoundType == EXCLUSIVE) {
        return other.isNumberInRange(lowerBound);
      }
      if (other.upperBoundType == EXCLUSIVE) {
        return isNumberInRange(other.lowerBound);
      }
    }

    return isNumberInRange(other.getLowerBound(), other.getLowerBoundType())
            && isNumberInRange(other.getUpperBound(), other.getUpperBoundType())
        || other.isNumberInRange(this.getLowerBound(), this.getLowerBoundType())
            && other.isNumberInRange(this.getUpperBound(), this.getUpperBoundType());
  }

  /**
   * Check if the current boundary subsumes the other boundary
   *
   * @param other boundary
   * @return true, if the current boundary subsumes the other boundary
   */
  public boolean subsumes(Boundary other) {
    if (isSubsumption(other)) {
      return isNumberInRange(other.getLowerBound(), other.getLowerBoundType())
              && isNumberInRange(other.getUpperBound(), other.getUpperBoundType())
          || lowerBound == other.lowerBound
              && (lowerBoundType == INCLUSIVE || lowerBoundType == other.lowerBoundType)
              && upperBound == other.upperBound
              && (upperBoundType == INCLUSIVE || upperBoundType == other.upperBoundType);
    }
    return false;
  }

  /**
   * Check, if a other Boundary is NOT equals and NOT overlapping and NO subsumption
   *
   * <p>Example:
   *
   * <blockquote>
   *
   * <pre>
   *    false: (overlapping)
   *     o------o
   *        o-----o
   *    true: (not overlapping)
   *     o---o
   *          o--o
   *    false: (subsumption)
   *      o--o
   *       o-o
   *    false: (equals)
   *      o---o
   *      o---o
   *    false: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  public boolean isNotInContact(Boundary other) {
    Objects.requireNonNull(other);
    return !this.isNumberInRange(other.getLowerBound(), other.getLowerBoundType())
        && !this.isNumberInRange(other.getUpperBound(), other.getUpperBoundType())
        && !other.isNumberInRange(this.getLowerBound(), this.getLowerBoundType())
        && !other.isNumberInRange(this.getUpperBound(), this.getUpperBoundType());
  }

  /**
   * Check, if a other Boundary is equals or overlapping or subsumption
   *
   * <p>Example:
   *
   * <blockquote>
   *
   * <pre>
   *    true: (overlapping)
   *     o------o
   *        o-----o
   *    false: (not overlapping)
   *     o---o
   *          o--o
   *    true: (subsumption)
   *      o--o
   *       o-o
   *    true: (equals)
   *      o---o
   *      o---o
   *    true: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  public boolean isInContact(Boundary other) {
    Objects.requireNonNull(other);
    return !isNotInContact(other);
  }

  /**
   * Check, if a other Boundary has equal bounds
   *
   * <p>Example:
   *
   * <blockquote>
   *
   * <pre>
   *    false: (overlapping)
   *     o------o
   *        o-----o
   *    false: (not overlapping)
   *     o---o
   *          o--o
   *    false: (subsumption)
   *      o--o
   *       o-o
   *    true: (equals)
   *      o---o
   *      o---o
   *    false: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  public boolean isEqualsTo(Boundary other) {
    return equals(other);
  }

  /**
   * Check, if a other Boundary has equal bounds
   *
   * <p>Example:
   *
   * <blockquote>
   *
   * <pre>
   *    false: (overlapping)
   *     o------o
   *        o-----o
   *    false: (not overlapping)
   *     o---o
   *          o--o
   *    false: (subsumption)
   *      o--o
   *       o-o
   *    true: (equals)
   *      o---o
   *      o---o
   *    false: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  @Override
  public boolean equals(Object other) {
    if (this == other) {
      return true;
    }
    if (other == null || getClass() != other.getClass()) {
      return false;
    }
    Boundary boundary = (Boundary) other;
    return Double.compare(boundary.lowerBound, lowerBound) == 0
        && Double.compare(boundary.upperBound, upperBound) == 0
        && lowerBoundType == boundary.lowerBoundType
        && upperBoundType == boundary.upperBoundType;
  }

  /**
   * Create an new boundary, which is lover than the current <br>
   * >10 --> <=10
   *
   * @return An Optional of the Boundary, if the current has a possible boundary, which is lower
   */
  public Optional<Boundary> getNewBoundaryLower() {
    if (lowerBound == Double.MIN_VALUE) {
      return Optional.empty();
    }
    return Optional.of(
        getBuilder()
            .withLowerBound(Double.MIN_VALUE)
            .withLowerBoundType(INCLUSIVE)
            .withUpperBound(lowerBound)
            .withUpperBoundType(lowerBoundType.getOp())
            .isInteger(isIntegerValue())
            .build());
  }

  /**
   * Create a new boundary between two existing ones<br>
   * <10 & >20 --> [10..20]
   *
   * @param other {@link Boundary}
   * @return An Optional of the new Boundary between the two given ones Return an empty Optional, if
   *     there is no interval in between
   */
  public Optional<Boundary> getNewBoundaryBetween(Boundary other) {
    if (isInContact(Objects.requireNonNull(other))) {
      return Optional.empty();
    }
    // not space between the two bounds
    if (this.upperBound == other.lowerBound && this.upperBoundType != other.lowerBoundType
        || other.upperBound == this.lowerBound && other.upperBoundType != this.lowerBoundType) {
      return Optional.empty();
    }
    // check if is integer val
    if (isIntegerValue()) {
      if (this.upperBound + 1 == other.lowerBound
              && this.upperBoundType == INCLUSIVE
              && other.lowerBoundType == INCLUSIVE
          || other.upperBound + 1 == this.lowerBound
              && other.upperBoundType == INCLUSIVE
              && this.lowerBoundType == INCLUSIVE) {
        return Optional.empty();
      }
    }
    if (this.upperBound < other.upperBound) {
      return Optional.of(
          getBuilder()
              .withLowerBound(this.upperBound)
              .withLowerBoundType(this.upperBoundType.getOp())
              .withUpperBound(other.lowerBound)
              .withUpperBoundType(other.lowerBoundType.getOp())
              .isInteger(isIntegerValue())
              .build());
    } else {
      return Optional.of(
          getBuilder()
              .withLowerBound(other.upperBound)
              .withLowerBoundType(other.upperBoundType.getOp())
              .withUpperBound(this.lowerBound)
              .withUpperBoundType(this.lowerBoundType.getOp())
              .isInteger(isIntegerValue())
              .build());
    }
  }

  /**
   * Create a new boundary, which begins at the upper bound of the original <br>
   * <10 --> >=10
   *
   * @return An Optional of the Boundary if the current has a possible boundary, which is higher
   */
  public Optional<Boundary> getNewBoundaryUpper() {
    if (upperBound == Double.MAX_VALUE) {
      return Optional.empty();
    }
    return Optional.of(
        getBuilder()
            .withLowerBound(upperBound)
            .withLowerBoundType(upperBoundType.getOp())
            .withUpperBound(Double.MAX_VALUE)
            .withUpperBoundType(INCLUSIVE)
            .isInteger(isIntegerValue())
            .build());
  }

  /**
   * Create a new Boundary from two given ones. The Optional is empty, if the given two bounds are
   * not direct followed by each other or current in contact<u>Example:</u><br>
   * <code>
   * [10..20]  &  >20 --> >=10<br> [10..20]  &  <10 --> <=20<br> [10..20]  &  <=0 --> empty Optional
   * (Boundaries are not in contact)<br> [10..20]  &  <=10 --> empty Optional (Boundaries are
   * overlapping)<br>
   * </code>
   *
   * @param other the other boundary
   */
  public Optional<Boundary> appendWith(Boundary other) {
    if (isNotInContact(other)) {
      if (this.upperBound == other.lowerBound && this.upperBoundType != other.lowerBoundType
          || isIntegerValue()
              && this.upperBound + 1 == other.lowerBound
              && this.upperBoundType == INCLUSIVE
              && other.lowerBoundType == INCLUSIVE) {
        return Optional.of(
            getBuilder()
                .withLowerBound(this.lowerBound)
                .withLowerBoundType(this.lowerBoundType)
                .withUpperBound(other.upperBound)
                .withUpperBoundType(other.getUpperBoundType())
                .isInteger(isIntegerValue())
                .build());
      } else if (this.lowerBound == other.upperBound && other.upperBoundType != this.lowerBoundType
          || isIntegerValue()
              && other.upperBound + 1 == this.lowerBound
              && other.upperBoundType == INCLUSIVE
              && this.lowerBoundType == INCLUSIVE) {
        return Optional.of(
            getBuilder()
                .withLowerBound(other.lowerBound)
                .withLowerBoundType(other.lowerBoundType)
                .withUpperBound(this.upperBound)
                .withUpperBoundType(this.getUpperBoundType())
                .isInteger(isIntegerValue())
                .build());
      }
    }
    return Optional.empty();
  }

  /**
   * Create a new Boundary from two given ones. The Optional is empty, if the given two bounds are
   * not direct followed by each other. They can be in contact <u>Example:</u><br>
   * <code>
   * [10..20]  &  >20 --> >=10<br> [10..20]  &  <10 --> <=20<br> [10..20]  &  <=0 --> empty Optional
   * (Boundaries are not in contact)<br> [10..20]  &  <=10 --> <=20 <br>
   * </code>
   *
   * @param other the other boundary
   */
  public Optional<Boundary> combineWith(Boundary other) {
    if (isNotInContact(other)) {
      return appendWith(other);
    }
    Builder builder = getBuilder().isInteger(isIntegerValue());
    if (this.lowerBound == Double.MIN_VALUE || this.lowerBound < other.lowerBound) {
      builder.withLowerBound(this.lowerBound).withLowerBoundType(this.lowerBoundType);
    } else if (other.lowerBound == Double.MIN_VALUE || other.lowerBound < this.lowerBound) {
      builder.withLowerBound(other.lowerBound).withLowerBoundType(other.lowerBoundType);
    } else {
      builder
          .withLowerBound(other.lowerBound)
          .withLowerBoundType(
              this.lowerBoundType == INCLUSIVE || other.lowerBoundType == INCLUSIVE
                  ? INCLUSIVE
                  : EXCLUSIVE);
    }
    if (this.upperBound == Double.MAX_VALUE || this.upperBound > other.upperBound) {
      builder.withUpperBound(this.upperBound).withUpperBoundType(this.upperBoundType);
    } else if (other.upperBound == Double.MAX_VALUE || other.upperBound > this.upperBound) {
      builder.withUpperBound(other.upperBound).withUpperBoundType(other.upperBoundType);
    } else {
      builder
          .withUpperBound(other.upperBound)
          .withUpperBoundType(
              this.upperBoundType == INCLUSIVE || other.upperBoundType == INCLUSIVE
                  ? INCLUSIVE
                  : EXCLUSIVE);
    }
    return Optional.of(builder.build());
  }

  @Override
  public int hashCode() {
    return Objects.hash(lowerBound, lowerBoundType, upperBound, upperBoundType);
  }

  /**
   * Parse a String to a boundary object.<br>
   * <u>Possible forms:</u><br>
   * <code>
   * Interval: [x..y] (inclusive x and inclusive y) <br> Interval: ]x..y[ (exclusive x and exclusive
   * y) <br> Max Val: <=x (inclusive x) <br> Max Val: &lt;x (exclusive x) <br> Min Val: >=x
   * (inclusive x) <br> Min Val: >x (exclusive x) <br>Equal: =x (inclusive x) <br>
   * </code>
   *
   * @param expr String to parse
   * @return a new Boundary object
   */
  public static Boundary parseFromNumericValue(String expr) {
    return parseFromNumericValue(expr, false);
  }

  /**
   * Parse a String to a boundary object.<br>
   * <u>Possible forms:</u><br>
   * <code>
   * Interval: [x..y] (inclusive x and inclusive y) <br> Interval: ]x..y[ (exclusive x and exclusive
   * y) <br> Max Val: <=x (inclusive x) <br> Max Val: &lt;x (exclusive x) <br> Min Val: >=x
   * (inclusive x) <br> Min Val: >x (exclusive x) <br>Equal: =x (inclusive x) <br>
   * </code>
   *
   * @param expr String to parse
   * @param isIntegerValue true, if the value is a integer Value
   * @return a new Boundary object
   */
  public static Boundary parseFromNumericValue(String expr, boolean isIntegerValue) {
    Builder builder = getBuilder();
    if (isIntegerValue) {
      builder.isInteger();
    }
    try {
      expr = Objects.requireNonNull(expr).trim();
      if (expr.startsWith("[") || expr.startsWith("]")) {
        builder
            .withLowerBoundType(expr.startsWith("[") ? INCLUSIVE : EXCLUSIVE)
            .withUpperBoundType(expr.endsWith("]") ? INCLUSIVE : EXCLUSIVE);
        expr = expr.replace('[', '\0').replace(']', '\0');
        String[] n = expr.split("\\.\\.");
        builder.withLowerBound(Double.parseDouble(n[0])).withUpperBound(Double.parseDouble(n[1]));
      } else if (expr.startsWith(">")) {
        builder.withLowerBoundType(expr.contains("=") ? INCLUSIVE : EXCLUSIVE);
        expr = expr.replace('=', '\0').replace('>', '\0');
        builder
            .withLowerBound(Double.parseDouble(expr))
            .withUpperBound(Double.MAX_VALUE)
            .withUpperBoundType(INCLUSIVE);
      } else if (expr.startsWith("<")) {
        builder.withUpperBoundType(expr.contains("=") ? INCLUSIVE : EXCLUSIVE);
        expr = expr.replace('=', '\0').replace('<', '\0');
        builder
            .withUpperBound(Double.parseDouble(expr))
            .withLowerBound(Double.MIN_VALUE)
            .withLowerBoundType(INCLUSIVE);
      } else if (expr.startsWith("=")) {
        expr = expr.replace('=', '\0');
        builder
            .withUpperBound(Double.parseDouble(expr))
            .withLowerBound(Double.parseDouble(expr))
            .withLowerBoundType(INCLUSIVE)
            .withUpperBoundType(INCLUSIVE);
      } else {
        throw new IllegalArgumentException();
      }
    } catch (Exception e) {
      throw new IllegalArgumentException(
          String.format("Parsing boundary failed: %s. Exception: %s", expr, e.getMessage()));
    }
    return builder.build();
  }

  /**
   * Parse a String value to a boundary object. The Value of the String equals the hashcode of the
   * string value and has the form<br>
   * <code>={stringValue.hashCode()}</code>
   *
   * @param stringValue the string to create the boundary
   * @return a boundary
   */
  public static Boundary parseFromStringValue(String stringValue) {
    int hash = Objects.requireNonNull(stringValue).hashCode();
    return getBuilder()
        .withUpperBound(hash)
        .withLowerBound(hash)
        .withLowerBoundType(INCLUSIVE)
        .withUpperBoundType(INCLUSIVE)
        .build();
  }

  /**
   * Parse a String value to a boundary object. The Value of the String equals the hashcode of the
   * string value and has the form<br>
   * <code>={stringValue.hashCode()}</code>
   *
   * @return a boundary
   */
  public static Boundary parseFromBooleanValue(boolean bolVal) {
    int val = bolVal ? 0 : 1;
    return getBuilder()
        .withUpperBound(val)
        .withLowerBound(val)
        .withLowerBoundType(INCLUSIVE)
        .withUpperBoundType(INCLUSIVE)
        .build();
  }

  /**
   * Creates a "wildcard" boundary. [-infinity..+infinity]
   *
   * @return a boundary
   */
  public static Boundary parseBoundaryFromNullValue() {
    // Wildcard
    return getBuilder()
        .withLowerBound(Double.MIN_VALUE)
        .withUpperBound(Double.MAX_VALUE)
        .withLowerBoundType(INCLUSIVE)
        .withUpperBoundType(INCLUSIVE)
        .build();
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    if (getLowerBound() == Double.MIN_VALUE && getUpperBound() == Double.MAX_VALUE) {
      // Wildcard
      sb.append("-");
    } else if (getLowerBound() == Double.MIN_VALUE) {
      sb.append("<");
      if (getUpperBoundType() == INCLUSIVE) {
        sb.append("=");
      }
      sb.append(getValAsString(getUpperBound()));
    } else if (getUpperBound() == Double.MAX_VALUE) {
      sb.append(">");
      if (getLowerBoundType() == INCLUSIVE) {
        sb.append("=");
      }
      sb.append(getValAsString(getLowerBound()));
    } else if (getLowerBound() == getUpperBound()) {
      sb.append("=").append(getValAsString(getLowerBound()));
    } else {
      if (getLowerBoundType() == INCLUSIVE) {
        sb.append("[");
      } else {
        sb.append("]");
      }
      sb.append(getValAsString(getLowerBound()));
      sb.append("..");
      sb.append(getValAsString(getUpperBound()));
      if (getUpperBoundType() == INCLUSIVE) {
        sb.append("]");
      } else {
        sb.append("[");
      }
    }
    return sb.toString();
  }

  private String getValAsString(double value) {
    if (isIntegerValue()) {
      return String.valueOf((new Double(value)).intValue());
    }
    return String.valueOf((new Double(value)));
  }

  @Override
  public int compareTo(Boundary o) {
    if (o == null) {
      return 1;
    }
    if (this.lowerBound == o.lowerBound) {
      if (this.lowerBoundType == o.lowerBoundType) {
        if (this.upperBound == o.upperBound) {
          if (this.upperBoundType == o.upperBoundType) {
            return 0;
          } else {
            return upperBoundType == EXCLUSIVE ? -1 : 1;
          }
        } else {
          if (this.upperBound == Double.MAX_VALUE) {
            return -1;
          } else if (o.upperBound == Double.MAX_VALUE) {
            return 1;
          }
        }
        return Double.compare(this.upperBound, o.upperBound);
      } else {
        return lowerBoundType == INCLUSIVE ? -1 : 1;
      }
    }
    if (this.lowerBound == Double.MIN_VALUE) {
      return -1;
    } else if (o.lowerBound == Double.MIN_VALUE) {
      return 1;
    } else {
      return Double.compare(this.lowerBound, o.lowerBound);
    }
  }

  public static Builder getBuilder() {
    return new Boundary().new Builder();
  }

  public class Builder extends DefaultBuilder<Boundary> {

    public Builder withUpperBound(double bound) {
      value.setUpperBound(bound);
      return this;
    }

    public Builder withLowerBound(double bound) {
      value.setLowerBound(bound);
      return this;
    }

    public Builder withUpperBoundType(BoundType boundType) {
      value.setUpperBoundType(boundType);
      return this;
    }

    public Builder withLowerBoundType(BoundType boundType) {
      value.setLowerBoundType(boundType);
      return this;
    }

    public Builder isInteger() {
      return isInteger(true);
    }

    public Builder isInteger(boolean val) {
      value.setIntegerValue(val);
      return this;
    }

    @Override
    protected void validate() {
      if (value.lowerBound == Double.MIN_VALUE) {
        value.lowerBoundType = INCLUSIVE;
      }
      if (value.upperBound == Double.MAX_VALUE) {
        value.upperBoundType = INCLUSIVE;
      }
      Objects.requireNonNull(value.lowerBoundType, "Lower bound type is null");
      Objects.requireNonNull(value.upperBoundType, "Upper bound type is null");
      if (value.lowerBound > value.upperBound && Double.MIN_VALUE != value.lowerBound) {
        throw new IllegalArgumentException(
            String.format(
                "Lower bound %f is greater than upper bound %f.",
                value.lowerBound, value.upperBound));
      }
      // if boundary values are declared as integer, than there should be no not int bounds
      if (value.isIntegerValue()) {
        if (value.lowerBound != Double.MIN_VALUE
            && value.lowerBound != Math.floor(value.lowerBound)) {
          throw new IllegalArgumentException(
              String.format("Lower bound %f is no integer/long value.", value.lowerBound));
        }
        if (value.upperBound != Double.MAX_VALUE
            && value.upperBound != Math.floor(value.upperBound)) {
          throw new IllegalArgumentException(
              String.format("Upper bound %f is no integer/long value.", value.upperBound));
        }
      }
    }
  }
}
