package de.unikoblenz.fgbks.dmn.core.verifier.helper;

import de.unikoblenz.fgbks.dmn.core.models.RuleIdentifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import org.camunda.bpm.dmn.engine.DmnDecision;
import org.camunda.bpm.dmn.engine.impl.DmnDecisionTableImpl;
import org.camunda.bpm.dmn.engine.impl.DmnDecisionTableInputImpl;
import org.camunda.bpm.dmn.engine.impl.DmnDecisionTableOutputImpl;
import org.camunda.bpm.dmn.engine.impl.DmnDecisionTableRuleImpl;
import org.camunda.bpm.dmn.engine.impl.DmnExpressionImpl;

/** Class for storing inputs and outputs from a DMNDecsion List */
public class RuleMap {

  private static final Logger LOGGER = Logger.getLogger(RuleMap.class.getSimpleName());
  private List<DmnDecision> dmnDecisions;
  private Map<InputType, List<Value>> inputs;
  private List<InputType> inputColumns;
  private Map<OutputType, List<Value>> outputs;
  private List<OutputType> outputColumns;
  private Set<String> dmnDecisionsKeys;

  private RuleMap(List<DmnDecision> dmnDecisions) {
    this.dmnDecisions = new ArrayList<>(dmnDecisions);
    this.inputs = new HashMap<>();
    this.inputColumns = new ArrayList<>();
    this.outputs = new HashMap<>();
    this.outputColumns = new ArrayList<>();
    this.dmnDecisionsKeys = new HashSet<>();
  }

  /**
   * Create a Map for all inputs and all outputs out of a list of DMN Decisions
   *
   * @param decisionList List of DMNDecision @NotNull
   * @return a {@link RuleMap} from the list of DMNDecision
   */
  public static RuleMap createMapFromDmn(List<DmnDecision> decisionList) {
    RuleMap ruleMap = new RuleMap(decisionList);
    for (DmnDecision d : decisionList) {
      if (d.isDecisionTable()) {
        ruleMap.dmnDecisionsKeys.add(d.getKey());
        DmnDecisionTableImpl table = (DmnDecisionTableImpl) d.getDecisionLogic();
        // inputs
        int i = 0;
        for (DmnDecisionTableInputImpl tableInput : table.getInputs()) {
          InputType it = new InputType(d, tableInput);
          ruleMap.inputColumns.add(it);
          ruleMap.inputs.put(it, new ArrayList<>());
          List<Value> inList = ruleMap.inputs.get(it);
          int r = 0;
          for (DmnDecisionTableRuleImpl rule : table.getRules()) {
            DmnExpressionImpl condition = rule.getConditions().get(i);
            inList.add(it.getValue(condition, d, ++r, rule.getId()));
          }
          i++;
        }
        // outputs
        i = 0;
        for (DmnDecisionTableOutputImpl tableOutput : table.getOutputs()) {
          OutputType ot = new OutputType(d, tableOutput);
          ruleMap.outputColumns.add(ot);
          ruleMap.outputs.put(ot, new ArrayList<>());
          List<Value> outList = ruleMap.outputs.get(ot);
          int r = 0;
          for (DmnDecisionTableRuleImpl rule : table.getRules()) {
            DmnExpressionImpl conclusion = rule.getConclusions().get(i);
            outList.add(ot.getValue(conclusion, d, ++r, rule.getId()));
          }
          i++;
        }
      }
    }
    return ruleMap;
  }

  /**
   * Return a list of all input values from a given type.
   *
   * @param type the given input type
   * @return a list of all Values from the type
   */
  public List<Value> getValuesFromInputType(Type type) {
    if (!inputs.keySet().contains(type)) {
      return Collections.emptyList();
    }
    return new ArrayList<>(inputs.get(type));
  }

  /**
   * Return a list of input values with the given type, filtered by the given rule identifiers
   *
   * @param type the given input type
   * @param ruleIdentifiersFilter the rules to filter
   * @return a list of given input types
   */
  public List<Value> getValuesFromInputType(Type type, List<RuleIdentifier> ruleIdentifiersFilter) {
    if (!inputs.keySet().contains(type)) {
      return Collections.emptyList();
    }
    if (ruleIdentifiersFilter == null) {
      return getValuesFromInputType(type);
    }
    return inputs
        .get(type)
        .stream()
        .filter(r -> ruleIdentifiersFilter.contains(r.getRuleIdentifier()))
        .collect(Collectors.toList());
  }

  /**
   * Return a list of all output values from a given type.
   *
   * @param type the given output type
   * @return a list of all values from the given output type
   */
  public List<Value> getValuesFromOutputType(Type type) {
    if (!outputs.keySet().contains(type)) {
      return Collections.emptyList();
    }
    return new ArrayList<>(outputs.get(type));
  }

  /**
   * Return a list of output values with the given type, filtered by the given rule identifiers
   *
   * @param type the given output type
   * @param ruleIdentifiersFilter the rules to filter
   * @return a list of all values from the given output type
   */
  public List<Value> getValuesFromOutputType(
      Type type, List<RuleIdentifier> ruleIdentifiersFilter) {
    if (!outputs.keySet().contains(type)) {
      return Collections.emptyList();
    }
    if (ruleIdentifiersFilter == null) {
      return getValuesFromOutputType(type);
    }
    return outputs
        .get(type)
        .stream()
        .filter(r -> ruleIdentifiersFilter.contains(r.getRuleIdentifier()))
        .collect(Collectors.toList());
  }

  /**
   * Return a list of output or input values with the given type
   *
   * @param type the given output or input type
   * @return a list of all values from the given output or input type
   */
  public List<Value> getValuesFromType(Type type) {
    if (outputs.keySet().contains(type)) {
      return new ArrayList<>(outputs.get(type));
    } else if (inputs.keySet().contains(type)) {
      return new ArrayList<>(inputs.get(type));
    }
    return Collections.emptyList();
  }

  /**
   * Return a list of output or input values with the given type, filtered by the given rule
   * identifiers
   *
   * @param type the given output or input type
   * @param ruleIdentifiersFilter the rules to filter
   * @return a list of all values from the given output or input type
   */
  public List<Value> getValuesFromType(Type type, List<RuleIdentifier> ruleIdentifiersFilter) {
    if (ruleIdentifiersFilter == null) {
      return getValuesFromType(type);
    }
    if (outputs.keySet().contains(type)) {
      return outputs
          .get(type)
          .stream()
          .filter(r -> ruleIdentifiersFilter.contains(r.getRuleIdentifier()))
          .collect(Collectors.toList());
    } else if (inputs.keySet().contains(type)) {
      return inputs
          .get(type)
          .stream()
          .filter(r -> ruleIdentifiersFilter.contains(r.getRuleIdentifier()))
          .collect(Collectors.toList());
    }
    return Collections.emptyList();
  }

  /**
   * Return a list of input types, which belongs to a given decisionKey
   *
   * @param decisionKey the decisionKey
   * @return a list of Input types
   */
  public List<Type> getAllInputTypesFromDecisionKey(String decisionKey) {
    return inputColumns
        .stream()
        .filter(i -> i.getDecisionKey().equals(Objects.requireNonNull(decisionKey)))
        .collect(Collectors.toList());
  }

  /**
   * Return a list of output types, which belongs to a given decisionKey
   *
   * @param decisionKey the decisionKey
   * @return a list of Output types
   */
  public List<Type> getAllOutputTypesFromDecisionKey(String decisionKey) {
    return outputColumns
        .stream()
        .filter(i -> i.getDecisionKey().equals(decisionKey))
        .collect(Collectors.toList());
  }

  /**
   * Return a Set of input and output types, which belongs to a given decisionKey
   *
   * @param decisionKey the decisionKey
   * @return a set of output and input types
   */
  public List<Type> getAllTypesFromDecisionKey(String decisionKey) {
    List<Type> types = getAllInputTypesFromDecisionKey(decisionKey);
    types.addAll(getAllOutputTypesFromDecisionKey(decisionKey));
    return types;
  }

  public Set<Set<String>> getIncludingInputDefinitions() {
    Set<Set<String>> returnSet = new HashSet<>();
    for (String d1 : dmnDecisionsKeys) {
      Set<String> subset = new HashSet<>();
      List<Type> inputsD1 = getAllInputTypesFromDecisionKey(d1);
      for (String d2 : dmnDecisionsKeys) {
        List<Type> inputsD2 = getAllInputTypesFromDecisionKey(d2);
        boolean containsAll = true;
        for (Type i1 : inputsD1) {
          boolean f2 = false;
          for (Type i2 : inputsD2) {
            if (i1.getExpression() != null && i1.getExpression().equals(i2.getExpression())) {
              f2 = true;
              break;
            }
          }
          if (!(containsAll &= f2)) {
            break;
          }
        }
        if (containsAll) {
          subset.add(d2);
        }
      }
      if (subset.size() > 1) {
        returnSet.add(subset);
      }
    }
    // check for subsets, that includes other subsets
    Iterator<Set<String>> it = returnSet.iterator();
    Set<Set<String>> removeSets = new HashSet<>();
    while (it.hasNext()) {
      Set<String> s = it.next();
      for (Set<String> subset : returnSet) {
        if (subset.size() > s.size() && subset.containsAll(s)) {
          removeSets.add(s);
        }
      }
    }
    returnSet.removeAll(removeSets);
    return returnSet;
  }

  public List<Value> getValuesFromInputExpression(Set<String> dmnDecisionsKeys, String expression) {
    Set<Type> types = new HashSet<>();
    for (Type type : inputs.keySet()) {
      if (type.getExpression().equals(expression)
          && dmnDecisionsKeys.contains(type.getDecisionKey())) {
        types.add(type);
      }
    }
    List<Value> valueList = new ArrayList<>();
    for (Type type : types) {
      valueList.addAll(getValuesFromInputType(type));
    }
    return valueList;
  }

  public List<Value> getValuesFromInputExpression(
      Set<String> dmnDecisionsKeys, String expression, List<RuleIdentifier> ruleIdentifiersFilter) {
    Set<Type> types = new HashSet<>();
    for (Type type : inputs.keySet()) {
      if (type.getExpression() != null
          && type.getExpression().equals(expression)
          && dmnDecisionsKeys.contains(type.getDecisionKey())) {
        types.add(type);
      }
    }
    List<Value> valueList = new ArrayList<>();
    for (Type type : types) {
      valueList.addAll(getValuesFromInputType(type, ruleIdentifiersFilter));
    }
    return valueList;
  }
}
