package de.unikoblenz.fgbks.dmn.core.verifier.helper;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import de.unikoblenz.fgbks.dmn.core.verifier.helper.Boundary.BoundType;
import org.junit.jupiter.api.Test;

class BoundaryTest {

  @Test
  void parseBoundaryRange() {
    Boundary b1 = Boundary.parseFromNumericValue("[1..10]");
    assertEquals(1, b1.getLowerBound());
    assertEquals(10, b1.getUpperBound());
    assertSame(b1.getLowerBoundType(), BoundType.INCLUSIVE);
    assertSame(b1.getUpperBoundType(), BoundType.INCLUSIVE);
    Boundary b2 = Boundary.parseFromNumericValue("]1 .. 10[");
    assertEquals(1, b2.getLowerBound());
    assertEquals(10, b2.getUpperBound());
  }

  @Test
  void parseBoundaryLowerHigher() {
    Boundary b1 = Boundary.parseFromNumericValue("<10");
    assertEquals(Double.MIN_VALUE, b1.getLowerBound());
    assertEquals(10, b1.getUpperBound());
    assertSame(b1.getUpperBoundType(), BoundType.EXCLUSIVE);
    Boundary b2 = Boundary.parseFromNumericValue(">10");
    assertEquals(10, b2.getLowerBound());
    assertEquals(Double.MAX_VALUE, b2.getUpperBound());
    assertSame(b1.getUpperBoundType(), BoundType.EXCLUSIVE);
  }

  @Test
  void parseBoundaryLowerHigherEquals() {
    Boundary b1 = Boundary.parseFromNumericValue("<=10");
    assertEquals(Double.MIN_VALUE, b1.getLowerBound());
    assertEquals(10, b1.getUpperBound());
    assertSame(b1.getUpperBoundType(), BoundType.INCLUSIVE);
    Boundary b2 = Boundary.parseFromNumericValue(">=10");
    assertEquals(10, b2.getLowerBound());
    assertEquals(Double.MAX_VALUE, b2.getUpperBound());
    assertSame(b1.getUpperBoundType(), BoundType.INCLUSIVE);
  }

  @Test
  void parseBoundaryEquals() {
    Boundary b1 = Boundary.parseFromNumericValue("=10");
    assertEquals(10, b1.getLowerBound());
    assertEquals(10, b1.getUpperBound());
    assertSame(b1.getUpperBoundType(), BoundType.INCLUSIVE);
    assertSame(b1.getLowerBoundType(), BoundType.INCLUSIVE);
  }

  @Test
  void parseBoundaryExceptions() {
    assertThrows(IllegalArgumentException.class, () -> Boundary.parseFromNumericValue("a<4"));
    assertThrows(IllegalArgumentException.class, () -> Boundary.parseFromNumericValue("[22.]"));
    assertThrows(IllegalArgumentException.class, () -> Boundary.parseFromNumericValue("[..]"));
    assertThrows(IllegalArgumentException.class, () -> Boundary.parseFromNumericValue("g"));
    assertThrows(IllegalArgumentException.class, () -> Boundary.parseFromNumericValue("4<"));
    assertThrows(IllegalArgumentException.class, () -> Boundary.parseFromNumericValue(">4,3"));
    assertThrows(IllegalArgumentException.class, () -> Boundary.parseFromNumericValue(">4,3"));
    assertThrows(IllegalArgumentException.class, () -> Boundary.parseFromNumericValue("[10..1]"));
  }

  @Test
  void numberInRange() {
    Boundary b1 = Boundary.parseFromNumericValue("=10");
    assertTrue(b1.isNumberInRange(10.0));
    assertFalse(b1.isNumberInRange(10.1));
    assertFalse(b1.isNumberInRange(9.9));

    b1 = Boundary.parseFromNumericValue("<10");
    assertFalse(b1.isNumberInRange(10.0));
    assertFalse(b1.isNumberInRange(10.1));
    assertTrue(b1.isNumberInRange(9.9));

    b1 = Boundary.parseFromNumericValue(">10");
    assertFalse(b1.isNumberInRange(10.0));
    assertTrue(b1.isNumberInRange(10.1));
    assertFalse(b1.isNumberInRange(9.9));

    b1 = Boundary.parseFromNumericValue(">=10");
    assertTrue(b1.isNumberInRange(10.0));
    assertTrue(b1.isNumberInRange(10.1));
    assertFalse(b1.isNumberInRange(9.9));

    b1 = Boundary.parseFromNumericValue("<=10");
    assertTrue(b1.isNumberInRange(10.0));
    assertFalse(b1.isNumberInRange(10.1));
    assertTrue(b1.isNumberInRange(9.9));

    b1 = Boundary.parseFromNumericValue("[1 .. 10]");
    assertFalse(b1.isNumberInRange(0.0));
    assertTrue(b1.isNumberInRange(1.0));
    assertTrue(b1.isNumberInRange(5.0));
    assertTrue(b1.isNumberInRange(10.0));
    assertFalse(b1.isNumberInRange(10.1));

    b1 = Boundary.parseFromNumericValue("]1 .. 10[");
    assertFalse(b1.isNumberInRange(0.0));
    assertFalse(b1.isNumberInRange(1.0));
    assertTrue(b1.isNumberInRange(5.0));
    assertFalse(b1.isNumberInRange(10.0));
    assertFalse(b1.isNumberInRange(10.1));

    b1 = Boundary.parseFromNumericValue("[1.5 .. 10[");
    assertTrue(b1.isNumberInRange(1.5));
    b1 = Boundary.parseFromNumericValue("<=1.5");
    assertTrue(b1.isNumberInRange(1.5));
  }

  @Test
  void notInContactTest() {
    Boundary b1 = Boundary.parseFromNumericValue("=10");
    Boundary b2 = Boundary.parseFromNumericValue("=10");
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("<10");
    b2 = Boundary.parseFromNumericValue("=10");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue(">10");
    b2 = Boundary.parseFromNumericValue("=10");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue(">10");
    b2 = Boundary.parseFromNumericValue("<10");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue(">10");
    b2 = Boundary.parseFromNumericValue("<10");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue(">=10");
    b2 = Boundary.parseFromNumericValue("<10");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue(">10");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue(">10");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue(">=10");
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10[");
    b2 = Boundary.parseFromNumericValue(">=10");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10[");
    b2 = Boundary.parseFromNumericValue(">10");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue("<=10");
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue("<10");
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("]0..10]");
    b2 = Boundary.parseFromNumericValue("<=0");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("]0..10]");
    b2 = Boundary.parseFromNumericValue("<0");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("]0..10]");
    b2 = Boundary.parseFromNumericValue("<5");
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("]0..10]");
    b2 = Boundary.parseFromNumericValue(">5");
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("]0..10]");
    b2 = Boundary.parseFromNumericValue("[-10..5]");
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue("[-10..0]");
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("]0..10]");
    b2 = Boundary.parseFromNumericValue("[-10..0]");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("]0..10]");
    b2 = Boundary.parseFromNumericValue("[-10..0[");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue("[-10..0[");
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue("[-10..100[");
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseBoundaryFromNullValue();
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("<=0", true);
    b2 = Boundary.parseFromNumericValue(">=0", true);
    assertTrue(b1.isInContact(b2));
    assertTrue(b2.isInContact(b1));
    assertFalse(b1.isNotInContact(b2));
    assertFalse(b2.isNotInContact(b1));

    b1 = Boundary.parseFromNumericValue("<0", true);
    b2 = Boundary.parseFromNumericValue(">0", true);
    assertFalse(b1.isInContact(b2));
    assertFalse(b2.isInContact(b1));
    assertTrue(b1.isNotInContact(b2));
    assertTrue(b2.isNotInContact(b1));
  }

  @Test
  void overlappingTest() {
    Boundary b1 = Boundary.parseFromNumericValue("=10");
    Boundary b2 = Boundary.parseFromNumericValue("=10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("<10");
    b2 = Boundary.parseFromNumericValue("=10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("<=10");
    b2 = Boundary.parseFromNumericValue("=10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("<10");
    b2 = Boundary.parseFromNumericValue("<9");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue(">10");
    b2 = Boundary.parseFromNumericValue(">9");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue(">=10");
    b2 = Boundary.parseFromNumericValue("=10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue(">10");
    b2 = Boundary.parseFromNumericValue("=10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue(">10");
    b2 = Boundary.parseFromNumericValue(">10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("<10");
    b2 = Boundary.parseFromNumericValue("<10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue("<=5");
    assertTrue(b1.isOverlapping(b2));
    assertTrue(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue(">5");
    assertTrue(b1.isOverlapping(b2));
    assertTrue(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue(">=5");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue("<10");
    assertTrue(b1.isOverlapping(b2));
    assertTrue(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue("<=10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]5..10[");
    b2 = Boundary.parseFromNumericValue("<5");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]5..10[");
    b2 = Boundary.parseFromNumericValue(">10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]5..10[");
    b2 = Boundary.parseFromNumericValue("<-10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]5..10[");
    b2 = Boundary.parseFromNumericValue(">-10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]5..10[");
    b2 = Boundary.parseFromNumericValue("<20");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10[");
    b2 = Boundary.parseFromNumericValue(">=0");
    assertTrue(b1.isOverlapping(b2));
    assertTrue(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10[");
    b2 = Boundary.parseFromNumericValue("<0");
    assertTrue(b1.isOverlapping(b2));
    assertTrue(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10[");
    b2 = Boundary.parseFromNumericValue("[0..10]");
    assertTrue(b1.isOverlapping(b2));
    assertTrue(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10[");
    b2 = Boundary.parseFromNumericValue("[10..10]");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10[");
    b2 = Boundary.parseFromNumericValue("]10..10]");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10]");
    b2 = Boundary.parseFromNumericValue("[10..20]");
    assertTrue(b1.isOverlapping(b2));
    assertTrue(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("[-10..10]");
    b2 = Boundary.parseFromNumericValue("[-20..-10]");
    assertTrue(b1.isOverlapping(b2));
    assertTrue(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue(">=5");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]5..10]");
    b2 = Boundary.parseFromNumericValue(">=5");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));

    b1 = Boundary.parseFromNumericValue("]5..10[");
    b2 = Boundary.parseFromNumericValue("<=10");
    assertFalse(b1.isOverlapping(b2));
    assertFalse(b2.isOverlapping(b1));
  }

  @Test
  void subsumptionTest() {
    Boundary b1 = Boundary.parseFromNumericValue("=10");
    Boundary b2 = Boundary.parseFromNumericValue("=10");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("<10");
    b2 = Boundary.parseFromNumericValue("=10");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("<=10");
    b2 = Boundary.parseFromNumericValue("=10");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("<10");
    b2 = Boundary.parseFromNumericValue("<9");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue(">10");
    b2 = Boundary.parseFromNumericValue(">9");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue(">=10");
    b2 = Boundary.parseFromNumericValue("=10");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue(">10");
    b2 = Boundary.parseFromNumericValue("=10");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue(">10");
    b2 = Boundary.parseFromNumericValue(">10");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue(">=10");
    b2 = Boundary.parseFromNumericValue(">=10");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("<10");
    b2 = Boundary.parseFromNumericValue("<10");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("<=10");
    b2 = Boundary.parseFromNumericValue("<=10");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue("<=5");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue(">5");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]5..10]");
    b2 = Boundary.parseFromNumericValue(">5");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue("<10");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue("<=10");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("[5..10]");
    b2 = Boundary.parseFromNumericValue(">=5");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]5..10]");
    b2 = Boundary.parseFromNumericValue(">=5");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]5..10]");
    b2 = Boundary.parseFromNumericValue(">5");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]5..10[");
    b2 = Boundary.parseFromNumericValue("<5");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]5..10[");
    b2 = Boundary.parseFromNumericValue(">10");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]5..10[");
    b2 = Boundary.parseFromNumericValue("<-10");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]5..10[");
    b2 = Boundary.parseFromNumericValue(">-10");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10[");
    b2 = Boundary.parseFromNumericValue(">=0");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10[");
    b2 = Boundary.parseFromNumericValue("<0");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10[");
    b2 = Boundary.parseFromNumericValue("[0..10]");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10]");
    b2 = Boundary.parseFromNumericValue("[0..10]");
    assertTrue(b1.isSubsumption(b2));
    assertTrue(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10[");
    b2 = Boundary.parseFromNumericValue("[10..10]");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10[");
    b2 = Boundary.parseFromNumericValue("]10..12]");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("]-10..10]");
    b2 = Boundary.parseFromNumericValue("[10..20]");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));

    b1 = Boundary.parseFromNumericValue("[-10..10]");
    b2 = Boundary.parseFromNumericValue("[-20..-10]");
    assertFalse(b1.isSubsumption(b2));
    assertFalse(b2.isSubsumption(b1));
  }

  @Test
  void getNewUpperBoundaryTest() {
    Boundary b1 = Boundary.parseFromNumericValue("=10");
    Boundary b2 = Boundary.parseFromNumericValue(">10");
    assertEquals(b2, b1.getNewBoundaryUpper().get());

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue(">10");
    assertEquals(b2, b1.getNewBoundaryUpper().get());

    b1 = Boundary.parseFromNumericValue("[0..10[");
    b2 = Boundary.parseFromNumericValue(">=10");
    assertEquals(b2, b1.getNewBoundaryUpper().get());

    b1 = Boundary.parseFromNumericValue(">10");
    assertFalse(b1.getNewBoundaryUpper().isPresent());

    b1 = Boundary.parseFromNumericValue(">=10");
    assertFalse(b1.getNewBoundaryUpper().isPresent());
  }

  @Test
  void getNewMiddleBoundaryTest() {
    Boundary b1 = Boundary.parseFromNumericValue(">10");
    Boundary b2 = Boundary.parseFromNumericValue("<10");
    Boundary b3 = Boundary.parseFromNumericValue("=10");
    assertEquals(b3, b1.getNewBoundaryBetween(b2).get());
    assertEquals(b3, b2.getNewBoundaryBetween(b1).get());

    b1 = Boundary.parseFromNumericValue(">20");
    b2 = Boundary.parseFromNumericValue("<10");
    b3 = Boundary.parseFromNumericValue("[10..20]");
    assertEquals(b3, b1.getNewBoundaryBetween(b2).get());
    assertEquals(b3, b2.getNewBoundaryBetween(b1).get());

    b1 = Boundary.parseFromNumericValue(">=20");
    b2 = Boundary.parseFromNumericValue("<=10");
    b3 = Boundary.parseFromNumericValue("]10..20[");
    assertEquals(b3, b1.getNewBoundaryBetween(b2).get());
    assertEquals(b3, b2.getNewBoundaryBetween(b1).get());

    b1 = Boundary.parseFromNumericValue(">=10");
    b2 = Boundary.parseFromNumericValue("<=10");
    assertFalse(b1.getNewBoundaryBetween(b2).isPresent());
    assertFalse(b2.getNewBoundaryBetween(b1).isPresent());

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue("[5..20]");
    assertFalse(b1.getNewBoundaryBetween(b2).isPresent());
    assertFalse(b2.getNewBoundaryBetween(b1).isPresent());
  }

  @Test
  void getNewLowerBoundaryTest() {
    Boundary b1 = Boundary.parseFromNumericValue("=10");
    Boundary b2 = Boundary.parseFromNumericValue("<10");
    assertEquals(b2, b1.getNewBoundaryLower().get());

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue("<0");
    assertEquals(b2, b1.getNewBoundaryLower().get());

    b1 = Boundary.parseFromNumericValue("]0..10[");
    b2 = Boundary.parseFromNumericValue("<=0");
    assertEquals(b2, b1.getNewBoundaryLower().get());

    b1 = Boundary.parseFromNumericValue("<10");
    assertFalse(b1.getNewBoundaryLower().isPresent());

    b1 = Boundary.parseFromNumericValue("<=10");
    assertFalse(b1.getNewBoundaryLower().isPresent());
  }

  @Test
  void combineBoundaryTest() {
    Boundary b1 = Boundary.parseFromNumericValue("=10");
    Boundary b2 = Boundary.parseFromNumericValue("<10");
    Boundary b3 = Boundary.parseFromNumericValue("<=10");
    assertEquals(b3, b1.appendWith(b2).get());
    assertEquals(b3, b2.appendWith(b1).get());

    b1 = Boundary.parseFromNumericValue("=10");
    b2 = Boundary.parseFromNumericValue(">10");
    b3 = Boundary.parseFromNumericValue(">=10");
    assertEquals(b3, b1.appendWith(b2).get());
    assertEquals(b3, b2.appendWith(b1).get());

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue(">10");
    b3 = Boundary.parseFromNumericValue(">=0");
    assertEquals(b3, b1.appendWith(b2).get());
    assertEquals(b3, b2.appendWith(b1).get());

    b1 = Boundary.parseFromNumericValue("[0..10]");
    b2 = Boundary.parseFromNumericValue("<0");
    b3 = Boundary.parseFromNumericValue("<=10");
    assertEquals(b3, b1.appendWith(b2).get());
    assertEquals(b3, b2.appendWith(b1).get());

    b1 = Boundary.parseFromNumericValue("]0..10[");
    b2 = Boundary.parseFromNumericValue(">=10");
    b3 = Boundary.parseFromNumericValue(">0");
    assertEquals(b3, b1.appendWith(b2).get());
    assertEquals(b3, b2.appendWith(b1).get());

    b1 = Boundary.parseFromNumericValue("[0..10[");
    b2 = Boundary.parseFromNumericValue("<0");
    b3 = Boundary.parseFromNumericValue("<10");
    assertEquals(b3, b1.appendWith(b2).get());
    assertEquals(b3, b2.appendWith(b1).get());
  }

  @Test
  void testIntegerVals() {
    Boundary b1 = Boundary.parseFromNumericValue("<0", true);
    Boundary b2 = Boundary.parseFromNumericValue("[5..10[", true);
    Boundary b3 = Boundary.parseFromNumericValue(">10", true);

    Boundary betw = Boundary.parseFromNumericValue("[0..5[", true);
    assertEquals(betw, b1.getNewBoundaryBetween(b2).get());
    assertEquals(betw, b2.getNewBoundaryBetween(b1).get());

    b1 = Boundary.parseFromNumericValue("]10..20]", true);
    b2 = Boundary.parseFromNumericValue("[5..10[", true);
    betw = Boundary.parseFromNumericValue("=10", true);
    assertEquals(betw, b1.getNewBoundaryBetween(b2).get());
    assertEquals(betw, b2.getNewBoundaryBetween(b1).get());

    b1 = Boundary.parseFromNumericValue(">100", true);
    b2 = Boundary.parseFromNumericValue("[0..99[", true);
    betw = Boundary.parseFromNumericValue("[99..100]", true);
    assertEquals(betw, b1.getNewBoundaryBetween(b2).get());
    assertEquals(betw, b2.getNewBoundaryBetween(b1).get());

    b1 = Boundary.parseFromNumericValue(">=100", true);
    b2 = Boundary.parseFromNumericValue("[0..99]", true);
    assertFalse(b1.getNewBoundaryBetween(b2).isPresent());
    assertFalse(b2.getNewBoundaryBetween(b1).isPresent());

    b1 = Boundary.parseFromNumericValue("[0..5[", true);
    b2 = Boundary.parseFromNumericValue("[5..10[", true);
    assertFalse(b1.getNewBoundaryBetween(b2).isPresent());
    assertFalse(b2.getNewBoundaryBetween(b1).isPresent());

    b1 = Boundary.parseFromNumericValue("=10", true);
    b2 = Boundary.parseFromNumericValue("=11", true);
    Boundary comb = Boundary.parseFromNumericValue("[10..11]", true);
    assertEquals(comb, b1.appendWith(b2).get());
    assertEquals(comb, b2.appendWith(b1).get());

    b1 = Boundary.parseFromNumericValue("[0..10]", true);
    b2 = Boundary.parseFromNumericValue("=11", true);
    comb = Boundary.parseFromNumericValue("[0..11]", true);
    assertEquals(comb, b1.appendWith(b2).get());
    assertEquals(comb, b2.appendWith(b1).get());

    b1 = Boundary.parseFromNumericValue("=10", true);
    b2 = Boundary.parseFromNumericValue("[11..20]", true);
    comb = Boundary.parseFromNumericValue("[10..20]", true);
    assertEquals(comb, b1.appendWith(b2).get());
    assertEquals(comb, b2.appendWith(b1).get());

    b1 = Boundary.parseFromNumericValue("=10", true);
    b2 = Boundary.parseFromNumericValue("]10..20]", true);
    comb = Boundary.parseFromNumericValue("[10..20]", true);
    assertEquals(comb, b1.appendWith(b2).get());
    assertEquals(comb, b2.appendWith(b1).get());
  }
}
